import numpy as np
from collections import defaultdict
import genetic_functions
np.seterr(all='ignore')


def networkx_topology(G):
    """Common Graphs
       nx.cycle_graph(n) is a ring topology
       nx.grid_2d_graph(n,n) nxn 2d lattice
       nx.hexagonal_lattice_graph(n,m) triangular lattices so that each "surrounded" node has 6 edges
       nx.connected_watts_strogatz_graph(n = 256, k = 10, p = .1) is supposed to model social networks"""

    node_to_index = {}
    for i, node in enumerate(G.nodes):
        node_to_index[node] = i

    topology_network = defaultdict(lambda: [])
    for edge in G.edges:
        index0 = node_to_index[edge[0]]
        index1 = node_to_index[edge[1]]
        topology_network[index0].append(index1)
        topology_network[index1].append(index0)
    topology_network = [(index, connections) for index, connections in topology_network.items()]
    topology_network = sorted(topology_network, key = lambda x: x[0])
    topology_network = [x[1] for x in topology_network]
    return topology_network


class GeneticOptimizer:
    def __init__(self,
                 objective,
                 G,
                 problem_type,
                 variables=None,
                 lb=None,
                 ub=None,
                 num_solutions=8,
                 mutation_rate=.01,
                 max_iterations=3000,
                 num_islands=128,
                 death_rate=.5,
                 min_target=None,
                 mutation_radius=0,
                 args=None,
                 integer_solution=False,
                 log_metrics=True):

        self.objective = objective
        self.G = G
        self.problem_type = problem_type
        self.lb = lb
        self.ub = ub
        self.variables = variables
        self.num_solutions = num_solutions
        self.mutation_rate = mutation_rate
        self.max_iterations = max_iterations
        self.num_islands = num_islands
        self.death_rate = death_rate
        self.min_target = min_target
        self.mutation_radius = mutation_radius
        self.args = args
        self.integer_solution = integer_solution
        if self.problem_type == 'real_valued':
            self.lb = lb.astype(np.float32)
            self.ub = ub.astype(np.float32)
            self.num_vars = len(self.lb)
        elif self.problem_type == 'sequence':
            self.num_vars = len(variables)
        self.log_metrics = log_metrics

        self.death_count = int(self.num_solutions*self.death_rate)
        self.topology_network = networkx_topology(self.G)

    def random_number_factory(self):
        size = min(self.max_iterations, 1000)
        other_island_indices = {island_index: (x for x in np.random.choice(self.topology_network[island_index],
                                                                           size=int((1-self.death_rate)*self.num_solutions*size),
                                                                           replace=True).astype(np.int32))
                                for island_index in range(self.num_islands)}

        # Compute parent indices ahead of time
        indices = [item for item in range(self.num_solutions-self.death_count)]
        parent_indices = np.random.choice(indices, size=(size, self.death_count*self.num_islands, 2), replace=True).astype(np.int32)

        # Compute mutation ahead of time
        mutations = {}
        random_uniforms = np.random.uniform(size=(size, self.num_islands*self.death_count, self.num_vars))
        mutations['needs_mutation'] = [np.array(np.where(x < self.mutation_rate)).T.astype(np.int32) for x in random_uniforms]
        mutations['lengths'] = [len(x) for x in mutations['needs_mutation']]
        mutations['random_mutation_var_indices'] = np.random.choice(range(self.num_vars), replace=True, size=sum(mutations['lengths'])).astype(np.int32)
        mutations['current_mutations_count'] = 0

        # Compute ordered_crossover_indices ahead of time
        if self.problem_type == 'real_valued':
            crossover_random_nums = np.random.uniform(low=0, high=1, size=(size, self.num_islands*self.death_count, self.num_vars)).astype(np.float32)
        elif self.problem_type == 'sequence':
            crossover_random_nums = np.sort(np.random.choice(range(self.num_vars), replace=True, size=(size, self.num_islands*self.death_count, 2))).astype(np.int32)
        return other_island_indices, parent_indices, crossover_random_nums, mutations

    def GA(self):
        if self.problem_type == 'sequence':
            island_solutions = [np.vstack([np.random.choice(self.variables, replace=False, size=len(self.variables))
                                           for _ in range(self.num_solutions)]).astype(np.int32)
                                for _ in range(self.num_islands)]
        if self.problem_type == 'real_valued':
            island_solutions = np.array([[[np.random.uniform(low=self.lb[var_index], high=self.ub[var_index])
                                           for var_index in range(self.num_vars)]
                                          for _ in range(self.num_solutions)]
                                         for __ in range(self.num_islands)]).astype(np.float32)
            if self.integer_solution:
                island_solutions = np.round(island_solutions).astype(np.int32)

        if self.args is None:
            island_fitnesses = [[self.objective(solution) for solution in solutions] for solutions in island_solutions]
            island_fitnesses = np.array(island_fitnesses).reshape((self.num_islands, self.num_solutions)).astype(np.float32)
        else:
            island_fitnesses = [[self.objective(solution, **self.args) for solution in solutions] for solutions in island_solutions]
            island_fitnesses = np.array(island_fitnesses).reshape((self.num_islands, self.num_solutions)).astype(np.float32)

        self.metrics = []
        iterations = 0
        while iterations < self.max_iterations:
            if iterations % 1000 == 0:
                # Should consider making a "random number factory" function.
                # Then alter on it could be adopted to take a stream or something clever.
                other_island_indices, parent_indices, crossover_random_nums, mutations = self.random_number_factory()

            # Order solutions by fitness
            argsorts = [np.argsort(island_fitnesses[island_index]) for island_index in range(self.num_islands)]
            island_solutions = [island_solutions[island_index][argsorts[island_index]] for island_index in range(self.num_islands)]
            island_fitnesses = [island_fitnesses[island_index][argsorts[island_index]] for island_index in range(self.num_islands)]

            # Natural selection: kill off worst self.death_rate of population
            island_solutions = [island_solutions[island_index][:-self.death_count] for island_index in range(self.num_islands)]
            island_fitnesses = [island_fitnesses[island_index][:-self.death_count] for island_index in range(self.num_islands)]

            # Choose parents
            parents = genetic_functions.random_selection(self.num_solutions, self.death_count, self.num_islands, other_island_indices, parent_indices[iterations % 1000])

            # Procreation and mutation
            L = mutations['lengths'][iterations % 1000]
            C = mutations['current_mutations_count']
            if self.problem_type == 'sequence':
                children = genetic_functions.procreate_sequence(parents=parents,
                                                                island_solutions=np.array(island_solutions).astype(np.int32),
                                                                ordered_crossover_indices=crossover_random_nums[iterations % 1000],
                                                                num_islands=self.num_islands,
                                                                death_count=self.death_count,
                                                                variables_len=len(self.variables))
                mutated_children = genetic_functions.swap_mutate(solutions=children,
                                                                 needs_mutation=mutations['needs_mutation'][iterations % 1000],
                                                                 random_var_indices=mutations['random_mutation_var_indices'][C:C+L])
            elif self.problem_type == 'real_valued':
                children = genetic_functions.procreate_real_valued(parents=parents,
                                                                   island_solutions=np.array(island_solutions).astype(np.float32),
                                                                   island_fitnesses=np.array(island_fitnesses).astype(np.float32),
                                                                   heuristic_crossover_randoms=crossover_random_nums[iterations % 1000],
                                                                   lb=self.lb,
                                                                   ub=self.ub,
                                                                   num_islands=self.num_islands,
                                                                   death_count=self.death_count,
                                                                   variables_len=self.num_vars)
                mutated_children = genetic_functions.mutate_random(solutions=children,
                                                                   lb=self.lb,
                                                                   ub=self.ub,
                                                                   needs_mutation=mutations['needs_mutation'][iterations % 1000],
                                                                   mutation_radius=self.mutation_radius)
                if self.integer_solution:
                    mutated_children = np.round(mutated_children).astype(np.int32)
            next_gen_solutions = mutated_children.reshape(self.num_islands, self.death_count, self.num_vars)
            mutations['current_mutations_count'] += L

            # Compute next gen fitness
            solutions_to_compute_fitness_for = []
            indices_for_computed_solutions = []
            [(solutions_to_compute_fitness_for.append(next_gen_solutions[island_index]),
              indices_for_computed_solutions.extend([island_index]*self.death_count))
             for island_index in range(self.num_islands)]
            if self.args is None:
                computed_fitnesses = [self.objective(solution) for solution in np.concatenate(solutions_to_compute_fitness_for)]  # self.objective(np.concatenate(solutions_to_compute_fitness_for))
            else:
                solutions_to_compute_fitness_for = np.concatenate(solutions_to_compute_fitness_for)
                computed_fitnesses = [self.objective(solution, **self.args)
                                      for solution in solutions_to_compute_fitness_for]

            # Apply next gen fitness
            next_gen_fitnesses = defaultdict(lambda: [])
            [next_gen_fitnesses[island_index].append(computed_fitnesses[index])
             for index, island_index in enumerate(indices_for_computed_solutions)]
            next_gen_fitnesses = [np.array(next_gen_fitnesses[island_index]) for island_index in range(self.num_islands)]

            # Concatenate survivors with next generation
            island_solutions = [np.concatenate((island_solutions[island_index], next_gen_solutions[island_index]))
                                for island_index in range(self.num_islands)]
            island_fitnesses = [np.concatenate((island_fitnesses[island_index], next_gen_fitnesses[island_index]))
                                for island_index in range(self.num_islands)]

            # Record results of generation
            if self.log_metrics:
                self.metrics.append((np.min(island_fitnesses), np.percentile(island_fitnesses, 25), np.percentile(island_fitnesses, 50)))
            iterations += 1
        all_fitnesses = np.vstack(island_fitnesses)
        ind = np.unravel_index(np.argmin(all_fitnesses, axis=None), all_fitnesses.shape)
        best_solution = island_solutions[ind[0]][ind[1]]
        best_fitness = all_fitnesses[ind]
        return (best_solution, best_fitness), self.metrics



from setuptools import setup, Extension, find_packages
from distutils.command.build_ext import build_ext

class CustomBuildExtCommand(build_ext):
    """build_ext command for use when numpy headers are needed."""
    def run(self):

        # Import numpy here, only when headers are needed
        import numpy

        # Add numpy headers to include_dirs
        self.include_dirs.append(numpy.get_include())

        # Call original build_ext command
        build_ext.run(self)
        
try:
    from Cython.Build import cythonize
except ImportError:
    use_cython = False
else:
    use_cython = True

cythonised_files = {}
ext_modules = []

if use_cython:
    ext_modules += [
        Extension(name="genetic_functions", sources=["darwyn/genopt/genetic_functions.pyx"]),
    ]
    cythonised_files = cythonize(ext_modules)
else:
    cythonised_files = [
        Extension(name="genetic_functions", sources=["darwyn/genopt/genetic_functions.c"]),
    ]

setup(
    name='darwyn-genopt',
    version='1.3.0',
    description='Optimizing analytic functions utilizing a genetic algorithm',
    url='https://github.com/kevinlioi/metaheuristics',
    author='Kevin Lioi',
    author_email='kevin@getdarwyn.com',
    license='unlicense',
    setup_requires=[
        # Setuptools 18.0 properly handles Cython extensions.
        'setuptools>=18.0',
        'cython',
        'numpy',
        'scipy',
        'numba',
        'pandas',
        'matplotlib',
        'line-profiler',
        'networkx'
    ],
    packages=find_packages(),
    ext_modules=cythonised_files,
    cmdclass = {'build_ext': CustomBuildExtCommand},
    install_requires=[
        'numpy',
        'networkx',
        'scipy',
        'numba'
    ],
    zip_safe=False
)



import numpy as np
import math
from numba import jit
from scipy.optimize import root
import bisect


@jit
def domination_test(x, score):
    """Returns the part of Array that is dominating score and the part that is nondominated with score.
       Also returns the indices that are non dominated with score"""

    first_better = x[:, 0] > score[0]
    second_better = x[:, 1] > score[1]
    dominating_count = (first_better & second_better).sum()
    non_dominated_with_score = first_better != second_better
    return dominating_count, x[non_dominated_with_score], non_dominated_with_score


class SimulatedAnnealing:
    def __init__(self,
                 objectives,
                 max_epochs=150,
                 M=1000,
                 a=.925,
                 diagnostic_sample_size=1000,
                 num_solutions_to_return=np.inf,
                 save_attainment_snapshots=False):
        """
            max_iterations: This shouldn't be changed (optimized default). It is the max_iterations of the simulated annealing procedure to do.
            M: It is the number of neighbors to find during one iteration of the simulated annealing procedure.
            a: It is the cooling schedule, set between 0 and 1. The higher it is, the slower the cooling.
        """

        self.objectives = objectives
        self.max_epochs = max_epochs
        self.M = M
        self.a = a
        self.diagnostic_sample_size = diagnostic_sample_size
        self.num_solutions_to_return = num_solutions_to_return
        self.save_attainment_snapshots = save_attainment_snapshots
        self.attainment_snapshots = []

        self.attainment_sampling_uniforms = np.random.uniform(size=10000)
        self.energy_uniforms = (x for x in np.random.uniform(size=10000))

        self.current_solution = self.objectives.create_random_solution()
        self.pareto_fitnesses = np.array([self.objectives.fitnesses(self.current_solution)]).astype(np.float32)
        self.pareto_fitnesses_sorted = [list(self.pareto_fitnesses[:, 0]), list(self.pareto_fitnesses[:, 1])]
        self.current_score = self.pareto_fitnesses[0]
        self.pareto_solutions = [self.current_solution]
        self.attainment_surface = self.attainment_sampling()
        self.as_dominating_current = domination_test(self.attainment_surface, self.current_score)[0]

    def optimize(self):
        """Uses simulated annealing to find a good answer but not the most optimal (since that is impossible)"""

        T_0 = self.diagnostics()

        temp_iteration = 0
        while temp_iteration < self.max_epochs:
            T_k = self.a**temp_iteration*T_0
            for m in range(self.M):
                self.multiobj_iteration(T_k)

            if self.save_attainment_snapshots:
                self.attainment_snapshots.append(self.attainment_surface.copy())
            temp_iteration += 1
        abridged_solutions, abridged_pareto_fitnesses = self.create_final_solutions()
        return abridged_solutions, abridged_pareto_fitnesses

    def diagnostics(self):
        """Does 100k iterations where a neighbor is always accepted. This helps define what T_0 should be,
           and sets up the pareto frontier before the real algorithm begins.

           T_0 should be chosen such that -sigma/ln(b) = T_0, where sigma is the sample variance of the change
           in energy and b is the percentage of answers to be accepted at the beginning of the simulated annealing procedure."""

        energies = []
        for i in range(self.diagnostic_sample_size):
            E = self.multiobj_iteration(0, diagnostics=True)
            energies.append(E)
        
        # Initial temperature is set so that 95% of random moves would be accepted
        get_initial_temp = lambda T_0: np.mean(np.exp(-(np.array([x for x in energies if x > 0]))/T_0)) - .95
        T_0 = root(get_initial_temp, 3, method='lm')['x'][0]  # lm method is more robust with the starting point than the hybr method
        return T_0

    def multiobj_iteration(self, T_k, diagnostics=False):
        new_solution = self.objectives.neighbor(self.current_solution)

        new_score = self.objectives.fitnesses(new_solution)

        as_dominating_new = domination_test(self.attainment_surface, new_score)[0]
        f_x_new = as_dominating_new
        f_x_current = self.as_dominating_current

        # Always need to see the dominance of current and new
        total = len(self.attainment_surface) + 2.0
        if (self.current_score[0] > new_score[0]) & (self.current_score[1] > new_score[1]):
            f_x_new += 1
        elif (new_score[0] > self.current_score[0]) & (new_score[1] > self.current_score[1]):
            f_x_current += 1

        # The more the attainment surface dominates a point, the less likely it is chosen.
        E = (f_x_new - f_x_current) / total
        if diagnostics:
            condition = True
        else:
            try:
                condition = next(self.energy_uniforms) < np.exp(-E/T_k)
            except StopIteration:
                self.energy_uniforms = (x for x in np.random.uniform(size=10000))
                condition = next(self.energy_uniforms) < np.exp(-E/T_k)

        if condition:
            # We are switching from current to next, so check if current needs to be added to archive
            # We need to add current if 1) current is at least non-dominated or 2) current dominates all of archive
            archive_dominating_current, non_dominated_archive, non_dominated_bool = domination_test(self.pareto_fitnesses, self.current_score)
            L1 = archive_dominating_current
            L2 = len(non_dominated_archive)
            if (L1 == 0) or (L2 == L1 == 0):
                indices_to_remove = np.where(non_dominated_bool == False)[0]
                fitnesses_to_remove_from_sort = []
                for index in sorted(indices_to_remove, reverse=True):
                    fitnesses_to_remove_from_sort.append(self.pareto_fitnesses[index])
                    del self.pareto_solutions[index]
                fitnesses_to_remove_from_sort = np.array(fitnesses_to_remove_from_sort)

                self.pareto_solutions.append(self.current_solution.copy())
                self.pareto_fitnesses = np.concatenate((non_dominated_archive, self.current_score[np.newaxis, :]))

                if len(fitnesses_to_remove_from_sort) > 0:
                    remove_indices0 = np.intersect1d(self.pareto_fitnesses_sorted[0], fitnesses_to_remove_from_sort[:, 0], assume_unique=True, return_indices=True)[1]
                    remove_indices1 = np.intersect1d(self.pareto_fitnesses_sorted[1], fitnesses_to_remove_from_sort[:, 1], assume_unique=True, return_indices=True)[1]
                    for index in sorted(remove_indices0, reverse=True):
                        del self.pareto_fitnesses_sorted[0][index]
                    for index in sorted(remove_indices1, reverse=True):
                        del self.pareto_fitnesses_sorted[1][index]
                index1 = bisect.bisect_left(self.pareto_fitnesses_sorted[0], self.current_score[0])
                index2 = bisect.bisect_left(self.pareto_fitnesses_sorted[1], self.current_score[1])
                self.pareto_fitnesses_sorted[0].insert(index1, self.current_score[0])
                self.pareto_fitnesses_sorted[1].insert(index2, self.current_score[1])

                self.attainment_surface = self.attainment_sampling()
            self.current_score = new_score
            self.current_solution = new_solution
            self.as_dominating_current = domination_test(self.attainment_surface, self.current_score)[0]
        if diagnostics:
            return E

    def attainment_sampling(self, num_pts=200):
        """Creates the Attainment Surface of the current pareto frontier (pareto_fitnesses). It is effectively a
           uniform distribution along the pareto frontier, but each point is dominated by exactly one pareto frontier
           point along one dimension."""

        L1, L2 = self.pareto_fitnesses_sorted
        L1_rev, L2_rev = L1[::-1], L2[::-1]
        L = len(L1)

        v1 = np.empty((num_pts, 2))
        v2 = np.empty((num_pts, 2))

        if len(self.attainment_sampling_uniforms) < 2*num_pts:
            self.attainment_sampling_uniforms = np.random.uniform(size=10000)
        L1_min, L1_max, L2_min, L2_max = L1[0], L1[-1], L2[0], L2[-1]
        v1[:, 0] = (L1_max - L1_min)*self.attainment_sampling_uniforms[:num_pts] + L1_min
        v2[:, 1] = (L2_max - L2_min)*self.attainment_sampling_uniforms[num_pts:2*num_pts] + L2_min
        self.attainment_sampling_uniforms = self.attainment_sampling_uniforms[2*num_pts:]

        Indices1 = -(L - np.searchsorted(L1, v1[:, 0]))
        Indices2 = -(L - np.searchsorted(L2, v2[:, 1]))

        v1[:, 1] = [L2_rev[index] for index in Indices1]
        v2[:, 0] = [L1_rev[index] for index in Indices2]
        attainment_surface = np.vstack((v1[:, :2], v2[:, :2]))
        return attainment_surface

    def create_final_solutions(self):
        """The ArchiveArray is typically too large to have an analyst determine which specific solution to use.
           This sorts the archive array according to the schedule/FTE metric and then selects about 50 solutions
           uniformly."""

        sort_indices = np.argsort(self.pareto_fitnesses[:, 1])
        abridged_pareto_fitnesses = []
        abridged_pareto_solutions = []
        modulus = math.ceil(len(self.pareto_fitnesses)/min(self.num_solutions_to_return, len(self.pareto_fitnesses)))
        for i, index in enumerate(sort_indices):
            if i % modulus == 0:
                abridged_pareto_fitnesses.append(self.pareto_fitnesses[index])
                abridged_pareto_solutions.append(self.pareto_solutions[index])
        abridged_pareto_fitnesses = np.vstack(abridged_pareto_fitnesses)
        return np.array(abridged_pareto_solutions), abridged_pareto_fitnesses


class MultiObjective:
    def __init__(self, lb, ub, objective_function, neighborhood_radius):
        self.lb = lb.astype(np.float32)
        self.ub = ub.astype(np.float32)
        self.objective_function = objective_function
        self.neighborhood_radius = neighborhood_radius

        self.R = self.neighborhood_radius**-1
        self.bound_range = self.ub - self.lb
        self.d = len(lb)
        self.adjustments = self.create_random_numbers()

    def create_random_numbers(self):
        # More efficient to do many random numbers at once rather than one-by-one
        u = np.random.normal(0, 1, (10000, self.d))
        norm = np.sqrt(np.sum(np.square(u), axis=1))
        r = np.random.uniform(size=10000)**(1.0/self.d)
        adjustments = (x for x in (r[:, np.newaxis]*u/norm[:, np.newaxis]) / self.R)
        return adjustments

    def create_random_solution(self):
        solution = np.array([np.random.uniform(low=lb, high=self.ub[var_index])
                             for var_index, lb in enumerate(self.lb)])
        return solution

    def neighbor(self, solution):
        normalized_solution = 2*((solution - self.lb) / self.bound_range) - 1
        while True:
            try:
                adjustment = next(self.adjustments)
            except StopIteration:
                self.adjustments = self.create_random_numbers()
                adjustment = next(self.adjustments)
            new_normalized_solution = normalized_solution + adjustment
            in_bounds = (-1 <= new_normalized_solution).all() & (new_normalized_solution <= 1).all()
            if in_bounds:
                break
        new_solution = self.bound_range*((new_normalized_solution + 1) / 2) + self.lb
        return new_solution

    def fitnesses(self, solution):
        return self.objective_function(solution)
